use super::Token;

#[derive(Debug, PartialEq)]
enum TokenType {
    Number(),
    Variable(),
    ImaginaryUnit(),
    Operator(),
    Substraction(),
    OpenParenthesis(),
    CloseParenthesis(),
    Equal(),
}

fn get_token_type(token: &Token) -> TokenType {
    match token {
        Token::Number(_) => TokenType::Number(),
        Token::Variable(_) => TokenType::Variable(),
        Token::ImaginaryUnit() => TokenType::ImaginaryUnit(),
        Token::OpenParenthesis() => TokenType::OpenParenthesis(),
        Token::CloseParenthesis() => TokenType::CloseParenthesis(),
        Token::Substraction() => TokenType::Substraction(),
        Token::Equal() => TokenType::Equal(),
        _ => TokenType::Operator(),
    }
}

pub fn sanitize_tokens(tokens: Vec<Token>) -> Result<Vec<Token>, String> {
    let mut sanitized_tokens: Vec<Token> = vec![];
    let mut last_token_type = None;
    let mut open_close_equal_count = (0, 0, 0);

    for token in tokens {
        let token_type = get_token_type(&token);
        match token_type {
            TokenType::OpenParenthesis() => open_close_equal_count.0 += 1,
            TokenType::CloseParenthesis() => {
                open_close_equal_count.1 += 1;
                if open_close_equal_count.1 > open_close_equal_count.0 {
                    return Err(format!("Error at token {:?}, closing parenthesis that was never opened", token));
                }
            },
            TokenType::Equal() => {
                open_close_equal_count.2 += 1;
                if open_close_equal_count.0 != open_close_equal_count.1 {
                    return Err(format!("Error, query has a different number of opening and closing parentheses"));
                }
                open_close_equal_count.0 = 0;
                open_close_equal_count.1 = 0;
            },
            _ => (),
        }
        if let Some(last_token_type) = last_token_type {
            match (&last_token_type, &token_type) {
                (TokenType::Number(), TokenType::Variable()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::Number(), TokenType::ImaginaryUnit()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::Number(), TokenType::OpenParenthesis()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::Variable(), TokenType::Number()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::Variable(), TokenType::ImaginaryUnit()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::Variable(), TokenType::OpenParenthesis()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::ImaginaryUnit(), TokenType::Number()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::ImaginaryUnit(), TokenType::Variable()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::ImaginaryUnit(), TokenType::OpenParenthesis()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::CloseParenthesis(), TokenType::OpenParenthesis()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::CloseParenthesis(), TokenType::Number()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::CloseParenthesis(), TokenType::Variable()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::CloseParenthesis(), TokenType::ImaginaryUnit()) => sanitized_tokens.push(Token::Multiplication()),
                (TokenType::OpenParenthesis(), TokenType::Operator()) => return Err(format!("Error at token {:?}, operator forbidden directly after opening parenthesis", token)),
                (TokenType::Operator(), TokenType::CloseParenthesis()) => return Err(format!("Error at token {:?}, all operators are forbidden directly before closing parenthesis", token)),
                (TokenType::Substraction(), TokenType::CloseParenthesis()) => return Err(format!("Error at token {:?}, all operators are forbidden directly before closing parenthesis", token)),
                (TokenType::OpenParenthesis(), TokenType::CloseParenthesis()) => return Err(format!("Error at token {:?}, empty parentheses", token)),
                (TokenType::Substraction(), TokenType::Equal()) => return Err(format!("Error at token {:?}, two incompatible tokens type in a row", token)),
                (TokenType::Substraction(), TokenType::Operator()) => return Err(format!("Error at token {:?}, two incompatible tokens type in a row", token)),
                (TokenType::Operator(), TokenType::Equal()) => return Err(format!("Error at token {:?}, two incompatible tokens type in a row", token)),
                (TokenType::Operator(), TokenType::Substraction()) => return Err(format!("Error at token {:?}, two incompatible tokens type in a row", token)),
                (TokenType::Equal(), TokenType::Operator()) => return Err(format!("Error at token {:?}, two incompatible tokens type in a row", token)),
                (TokenType::OpenParenthesis(), TokenType::OpenParenthesis()) => (),
                (TokenType::CloseParenthesis(), TokenType::CloseParenthesis()) => (),
                _ => {
                    if token_type == last_token_type {
                        return Err(format!("Error at token {:?}, two incompatible tokens type in a row", token));
                    }
                },
            }
        } else if token_type == TokenType::Operator() || token_type == TokenType::Equal() {
            return Err(format!("Error at token {:?}, query can't start with this token", token));
        }
        sanitized_tokens.push(token);

        last_token_type = Some(token_type);
    }
    if open_close_equal_count.0 != open_close_equal_count.1 {
        return Err(format!("Error, query has a different number of opening and closing parentheses"));
    }
    if open_close_equal_count.2 > 1 {
        return Err(format!("Error, query shouldn't have more than 1 equals sign"));
    }
    let last_token = sanitized_tokens.last().unwrap();
    match get_token_type(&last_token) {
        TokenType::Equal() | TokenType::Operator() | TokenType::Substraction() => return Err(format!("Error at token {:?}, query can't end with this token", last_token)),
        _ => (),
    }
    Ok(sanitized_tokens)
}

#[cfg(test)]
mod tests {
    use super::*;

    fn plus() -> Token {
        Token::Addition()
    }

    fn minus() -> Token {
        Token::Substraction()
    }

    fn times() -> Token {
        Token::Multiplication()
    }

    fn divided_by() -> Token {
        Token::Division()
    }

    fn modulo() -> Token {
        Token::Modulo()
    }

    fn equals() -> Token {
        Token::Equal()
    }

    fn x() -> Token {
        Token::Variable(String::from("x"))
    }

    fn i() -> Token {
        Token::ImaginaryUnit()
    }

    fn number() -> Token {
        Token::Number(String::from("123"))
    }

    fn open() -> Token {
        Token::OpenParenthesis()
    }

    fn close() -> Token {
        Token::CloseParenthesis()
    }

    #[test]
    #[should_panic]
    fn double_number() {
        let tokens = vec![number(), number()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn double_operator() {
        let tokens = vec![number(), plus(), divided_by()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn double_variable() {
        let tokens = vec![x(), x()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn double_imaginary_unit() {
        let tokens = vec![i(), i()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn sanitize_tokens_starts_with_operator() {
        let tokens = vec![plus(), number(), divided_by(), x()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    fn starts_with_negative_number() {
        let tokens = vec![minus(), number(), divided_by(), x()];
        let results = tokens.clone();
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    #[test]
    fn number_operator_variable() {
        let tokens = vec![number(), divided_by(), x()];
        let results = tokens.clone();
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    #[test]
    fn insert_multiplication() {
        let tokens = vec![number(), x()];
        let results = vec![number(), times(), x()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![x(), number()];
        let results = vec![x(), times(), number()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![number(), i()];
        let results = vec![number(), times(), i()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![i(), number()];
        let results = vec![i(), times(), number()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![i(), x()];
        let results = vec![i(), times(), x()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![x(), i()];
        let results = vec![x(), times(), i()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    #[test]
    fn insert_multiplication_parenthesis() {
        let tokens = vec![open(), number(), modulo(), number(), close(), open(), i(), times(), x(), close()];
        let results = vec![open(), number(), modulo(), number(), close(), times(), open(), i(), times(), x(), close()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![number(), open(), number(), modulo(), number(), close()];
        let results= vec![number(), times(), open(), number(), modulo(), number(), close()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![i(), open(), number(), modulo(), number(), close()];
        let results= vec![i(), times(), open(), number(), modulo(), number(), close()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![x(), open(), number(), modulo(), number(), close()];
        let results= vec![x(), times(), open(), number(), modulo(), number(), close()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![open(), number(), modulo(), number(), close(), number()];
        let results= vec![open(), number(), modulo(), number(), close(), times(), number()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![open(), number(), modulo(), number(), close(), i()];
        let results= vec![open(), number(), modulo(), number(), close(), times(), i()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![open(), number(), modulo(), number(), close(), x()];
        let results= vec![open(), number(), modulo(), number(), close(), times(), x()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec![number(), open(), number(), modulo(), number(), close(), x()];
        let results= vec![number(), times(), open(), number(), modulo(), number(), close(), times(), x()];
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    #[test]
    #[should_panic]
    fn parenthesis_operator() {
        let tokens = vec!(open(), plus(), number(), close());
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn operator_parenthesis() {
        let tokens = vec!(open(), number(), modulo(), close());
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn minus_parenthesis() {
        let tokens = vec!(open(), number(), minus(), close());
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn empty_parentheses() {
        let tokens = vec!(open(), close());
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    fn parenthesis_minus() {
        let tokens = vec!(open(), minus(), number(), close());
        let results = tokens.clone();
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    #[test]
    fn double_parentheses() {
        let tokens = vec!(open(), open(), number(), plus(), number(), close(), plus(), number(), close());
        let results = tokens.clone();
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec!(open(), number(), plus(), open(), number(), plus(), number(), close(), close());
        let results = tokens.clone();
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec!(open(), open(), number(), plus(), number(), close(), number(), close());
        let results = vec!(open(), open(), number(), plus(), number(), close(), times(), number(), close());
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);

        let tokens = vec!(open(), number(), open(), number(), plus(), number(), close(), close());
        let results= vec!(open(), number(), times(), open(), number(), plus(), number(), close(), close());
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    #[test]
    #[should_panic]
    fn wrong_parentheses_count() {
        let tokens = vec![open(), number(), close(), close()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn wrong_parentheses_count_2() {
        let tokens = vec![open(), open(), number(), close()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn wrong_parentheses_count_hard() {
        let tokens = vec![open(), open(), number(), close(), equals(), x(), close()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn wrong_equal_count() {
        let tokens = vec![number(), equals(), number(), equals(), number()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn starts_with_equal() {
        let tokens = vec![equals(), number(), plus(), number()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn ends_with_equal() {
        let tokens = vec![number(), plus(), number(), equals()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn ends_with_minus() {
        let tokens = vec![number(), plus(), number(), minus()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn ends_with_operator() {
        let tokens = vec![number(), plus(), number(), divided_by()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn ends_with_open_brackets() {
        let tokens = vec![number(), plus(), number(), open(), number(), close(), open()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn starts_with_close_brackets() {
        let tokens = vec![close(), number(), plus(), number(), open(), number()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn wrong_brackets_order() {
        let tokens = vec![number(), close(), number(), plus(), number(), open(), number()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn minus_operator() {
        let tokens = vec![number(), minus(), plus(), i()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn operator_minus() {
        let tokens = vec![number(), plus(), minus(), i()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn minus_equals() {
        let tokens = vec![number(), minus(), equals(), i()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn operator_equals() {
        let tokens = vec![number(), modulo(), equals(), i()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    #[should_panic]
    fn equals_operator() {
        let tokens = vec![number(), equals(), times(), i()];
        sanitize_tokens(tokens).unwrap();
    }

    #[test]
    fn equals_minus() {
        let tokens = vec![number(), equals(), minus(), x()];
        let results = tokens.clone();
        assert_eq!(sanitize_tokens(tokens).unwrap(), results);
    }

    mod get_token_type {
        use super::*;

        #[test]
        fn exhaustive() {
            assert_eq!(get_token_type(&Token::Number(String::from("123"))), TokenType::Number());
            assert_eq!(get_token_type(&Token::Variable(String::from("var"))), TokenType::Variable());
            assert_eq!(get_token_type(&Token::ImaginaryUnit()), TokenType::ImaginaryUnit());
            assert_eq!(get_token_type(&Token::OpenParenthesis()), TokenType::OpenParenthesis());
            assert_eq!(get_token_type(&Token::CloseParenthesis()), TokenType::CloseParenthesis());
            assert_eq!(get_token_type(&Token::Substraction()), TokenType::Substraction());
            assert_eq!(get_token_type(&Token::Equal()), TokenType::Equal());
            assert_eq!(get_token_type(&Token::Addition()), TokenType::Operator());
            assert_eq!(get_token_type(&Token::Multiplication()), TokenType::Operator());
            assert_eq!(get_token_type(&Token::Division()), TokenType::Operator());
            assert_eq!(get_token_type(&Token::Modulo()), TokenType::Operator());
            assert_eq!(get_token_type(&Token::Exponentiation()), TokenType::Operator());
        }
    }
}
